/*
 * Copyright 2015 bvtest <email>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */


#include "mainwindow.h"
#include "cdeviceselectiondlg.h"
#include "cpropertiesdialog.h"
#include <QSizePolicy>


/////////////////////////////////////////////////////////////////////
// Constructor of the MainWindow. Besides of QT5 elements, the
// TcamCamera object in _pCam is created and the barcode scanner
// is configured.
//
MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent)
{
    // Instance of our TcamCamera class
    _pCam = NULL;

    window = new QWidget(this);

    setCentralWidget( window );
    mainLayout = new QHBoxLayout;
    window->setLayout(mainLayout);


    // Create the file menu
    _fileMenu = menuBar()->addMenu(tr("&File"));

    _exitAct = new QAction(tr("&Exit"), this);
    _exitAct->setShortcuts(QKeySequence::Close);
    _exitAct->setStatusTip(tr("Exit program"));
    connect(_exitAct, &QAction::triggered, this, &QWidget::close);
    _fileMenu->addAction(_exitAct);


    // Create the Device menu
    _deviceMenu = menuBar()->addMenu(tr("&Device"));

    _DeviceSelectAct = new QAction(tr("&Select"), this);
    _DeviceSelectAct->setStatusTip(tr("Select and start a device"));
    connect(_DeviceSelectAct, &QAction::triggered, this, &MainWindow::OnSelectDevice);
    _deviceMenu->addAction(_DeviceSelectAct);

    _DevicePropertiesAct = new QAction(tr("&Properties"), this);
    _DevicePropertiesAct->setStatusTip(tr("Show device property dialog"));
    connect(_DevicePropertiesAct, &QAction::triggered, this, &MainWindow::OnDeviceProperties);
    _deviceMenu->addAction(_DevicePropertiesAct);


    QHBoxLayout *HLayout = new QHBoxLayout;
    // Create the video display Widget
    _VideoWidget = new QWidget(NULL);
    _VideoWidget->setAttribute(Qt::WA_PaintOnScreen);
	_VideoWidget->setAttribute(Qt::WA_OpaquePaintEvent);
    _VideoWidget->setMinimumSize(640,480);

    // The line edit will list found barcodes
    _BarcodeList = new QTextEdit();
    _BarcodeList->setMinimumSize(100,480);
    _BarcodeList->setAlignment(Qt::AlignTop);

    _BarcodeList->setSizePolicy(QSizePolicy::Fixed,QSizePolicy::Expanding );

    mainLayout->addWidget(_VideoWidget);
    mainLayout->addWidget(_BarcodeList);
    
  
    // Instantiate the barcode scanner.
    _CustomData.pIC_BarcodeScanner =  ICBarcode_CreateScanner();
    _CustomData.BarCodes.resize(50);

    // Configure the barcode scanner to try harder. If barcodes are hardly detected,
    // pass an 1 instead of a 0.
    ICBarcode_SetParam(_CustomData.pIC_BarcodeScanner, IC_BARCODEPARAMS_TRY_HARDER, 0);

    // Enable barcode checksum support. 
	ICBarcode_SetParam(_CustomData.pIC_BarcodeScanner, IC_BARCODEPARAMS_CHECK_CHAR, 1);

    // Define. which formats the barcode library shall detect.

	int formats = 0;
	formats |= ICBarcode_Format::IC_BARCODEFORMAT_CODE_128;
	formats |= ICBarcode_Format::IC_BARCODEFORMAT_CODE_93;
	formats |= ICBarcode_Format::IC_BARCODEFORMAT_EAN_13;
	formats |= ICBarcode_Format::IC_BARCODEFORMAT_EAN_8;
	formats |= ICBarcode_Format::IC_BARCODEFORMAT_UPC_A;
	formats |= ICBarcode_Format::IC_BARCODEFORMAT_QR_CODE;
	//formats |= ICBarcode_Format::IC_BARCODEFORMAT_INTERLEAVED_2_OF_5;
	//formats |= ICBarcode_Format::IC_BARCODEFORMAT_CODE_39;
    
	ICBarcode_SetBarcodeFormats(_CustomData.pIC_BarcodeScanner, formats);

    // Try to load the last used camera. If that fails, show the device selection dialog.
    if( !LoadLastUsedDevice() )
    {
        OnSelectDevice();
    }
}


////////////////////////////////////////////////////////////////////////
// Destuctor. Stop the live stream and delete the TCamCamera object.
//
MainWindow::~MainWindow()
{
    if( _pCam != NULL)
    {
        _pCam->stop();
        delete _pCam;
    }
}

///////////////////////////////////////////////////////////////////////
// Handler for camera or other capture device selection.
//
void MainWindow::OnSelectDevice()
{

    // Create the device selection dialog
    CDeviceSelectionDlg Dlg(window );
    if( Dlg.exec() == 1 )
    {

        OpenandStartDevice(Dlg.getSerialNumber(),
                           Dlg.getWidth(),Dlg.getHeight(),
                           Dlg.getFPSNominator(),Dlg.getFPSDeNominator());

        SaveLastUsedDevice(Dlg.getSerialNumber(),
                           Dlg.getWidth(),Dlg.getHeight(),
                           Dlg.getFPSNominator(),Dlg.getFPSDeNominator());
    }
}

/////////////////////////////////////////////////////////////////////
// Handler for showing the device settings dialog for exposure, gain etc.
//
void MainWindow::OnDeviceProperties()
{
    if( _pCam != NULL )
    {
        CPropertiesDialog Properties(nullptr);
        TcamProp* ptcam = (TcamProp*)_pCam->getTcamBin();
        // Pass the tcambin element to the properties dialog
        // so in knows, which device do handle
        Properties.SetCamera(ptcam);
        Properties.exec();
    }
}

//////////////////////////////////////////////////////////////////////////////
// Open the camea set video format and fps and start.
//
void MainWindow::OpenandStartDevice(std::string serial, int width, int height, int fps1, int fps2)
{
        if( _pCam != NULL)
        {
            _pCam->stop();
            delete _pCam;
        }
        // Instantiate the TcamCamera object with the serial number
        // of the selected device
        _pCam = new gsttcam::TcamCamera(serial);

        // Set video format, resolution and frame rate. We display color.
        _pCam->set_capture_format("BGRx", 
                            gsttcam::FrameSize{width,height}, 
                            gsttcam::FrameRate{fps1,fps2});

        // Get the window handle of the video widget
        WId xwinid = _VideoWidget->winId();

        // Create the display sink
        GstElement *imagesink = gst_element_factory_make("ximagesink", NULL);

        // Assing the window handle
        gst_video_overlay_set_window_handle(GST_VIDEO_OVERLAY(imagesink) , xwinid);

        // Pass the display sink to the TcamCamera object
        _pCam->enable_video_display(imagesink);

        _CustomData.ImageCounter = 0;
        _CustomData.busy = false;
        _CustomData.pMainWindow = this;

        _pCam->set_new_frame_callback(MainWindow::new_frame_cb, &_CustomData );

        // Show the live vieo
        _pCam->start();
}

///////////////////////////////////////////////////////////////////////////////////
// Save the last used device. At least, ist is only the serial number, video format's
// width and height and the frame rate. Propertiers are not saved.
//
void MainWindow::SaveLastUsedDevice(std::string serial, int width, int height, int fps1, int fps2)
{
    FILE *pFile = fopen("lastuseddevice.txt", "w");
    if( pFile )
    {
        fprintf(pFile,"%s\n%d\n%d\n%d\n%d\n", serial.c_str(), width,height,fps1,fps2);
        fclose( pFile);
    }
}
////////////////////////////////////////////////////////////////////////////////
// Restore the last used device. Only video format's width and height and the frame
// rate is restored.
//
bool MainWindow::LoadLastUsedDevice()
{
    char serial[50];
    int width;
    int height;
    int fps1;
    int fps2;

    FILE *pFile = fopen("lastuseddevice.txt", "r");
    if( pFile )
    {
        fscanf(pFile,"%s", serial );
        fscanf(pFile,"%d", &width);
        fscanf(pFile,"%d", &height);
        fscanf(pFile,"%d", &fps1);
        fscanf(pFile,"%d", &fps2);
        fclose( pFile);

        OpenandStartDevice( serial,  width,  height,  fps1,  fps2);
        return true;
    }

    return false;
}

//////////////////////////////////////////////////////////////////////////////////////////
// Event handler, so we can call MainWindow functions from the camera thread.
void MainWindow::postBarcodeFoundEvent(const int customData1, const int customData2)
{   
    // This method (postMyCustomEvent) can be called from any thread

    QApplication::postEvent(this, new BarcodeFoundEvent(customData1, customData2));   
}

void MainWindow::customEvent(QEvent * event)
{
    // When we get here, we've crossed the thread boundary and are now
    // executing in the Qt object's thread

    if(event->type() == BARCODE_FOUND_EVENT)
    {
        handleBarcodeFoundEvent(static_cast<BarcodeFoundEvent *>(event));
    }

    // use more else ifs to handle other custom events
}

///////////////////////////////////////////////////////////////////////////////
// This event is called from the GStreamer callback of TCamcamera class.
// Barcodes are already deceted and they are in the _CustomData.FoundBarcodes
// member.
// The "svg" string will contain the xml string for the rectangles around the
// found barcodes.
//
void MainWindow::handleBarcodeFoundEvent(const BarcodeFoundEvent *event)
{
    std::string BarcodeList = "";
    std::string svg;
    char primitive[1024];
    svg = "<svg>\n";


    for( int i = 0 ; i < _CustomData.FoundBarcodes; i++)
    {
        BarcodeList += _CustomData.BarCodes[i].Text; // Get the contents of a decoded barcode
        BarcodeList += "\n";

        // Create a polygon using the x/y coordinates of the corner of a barcode in the image
        svg += "\t<polygon points=\"";
        for(int p = 0; p < _CustomData.BarCodes[i].ResultPointsLength; p++)
        {
            if( p > 0)
            {
                svg += ",";
            }
            sprintf(primitive,"%d,%d",
                (int)_CustomData.BarCodes[i].ResultPoints[p].x,
                (int)_CustomData.BarCodes[i].ResultPoints[p].y);
            svg += primitive;
        }
        svg += "\" fill='none' stroke='red' stroke-width='3px'/>\n";
    }
    svg += "</svg>\n";
    _pCam->SetOverlayGraphicString(svg.c_str()); // Pass the graphics string to the overly
    _BarcodeList->setText(BarcodeList.c_str());  // Show the decoded barcodes in the barcode window.
}


////////////////////////////////////////////////////////////////////
// Callback called for new images by the internal appsink
// In the sample the GST Buffer properties, such as width and height and
// pixel format are retrieved. 
// Optionally the buffer is convertet to Grayscale, then the 
// barcode scanning is done.
// After barcode scanning, the event for displaying the results in the
// MainWindow class is called.
// 
// The pointer data is casted to CUSTOMDATA, so the found barcodes can 
// be saved for use in MainWindow.
//
GstFlowReturn MainWindow::new_frame_cb(GstAppSink *appsink, gpointer data)
{
    int width, height ;
    const GstStructure *str;

    GstSample *sample = gst_app_sink_pull_sample(appsink);

    // Cast gpointer to CUSTOMDATA*
    CUSTOMDATA *pCustomData = (CUSTOMDATA*)data;

    // Skip this frame, in case we are still busy.
    if( !pCustomData->busy )
    {
        pCustomData->busy = true;

        pCustomData->ImageCounter++;

        GstBuffer *buffer = gst_sample_get_buffer(sample);

        GstMapInfo info;

        gst_buffer_map(buffer, &info, GST_MAP_READ);
        
        if (info.data != NULL) 
        {
            // info.data contains the image data as blob of unsigned char 

            GstCaps *caps = gst_sample_get_caps(sample);
            // Get a string containg the pixel format, width and height of the image        
            str = gst_caps_get_structure (caps, 0);    
                // Now query the width and height of the image
            gst_structure_get_int (str, "width", &width);
            gst_structure_get_int (str, "height", &height);

            if( strcmp( gst_structure_get_string (str, "format"),"BGRx") == 0)  
            {
                // If the buffer is a 32 bit RGB32 image, it must be converted to
                // 8 bit grayscale. before barcodes can be searched.

                std::vector<unsigned char> buffer;
                buffer.resize(width * height);    
                ICBarcode_Transform_BGRA32_to_Y800(
                    buffer.data(), info.data,	
                    width, height,
                    width * 4,
                    width
                    );
                
                pCustomData->FoundBarcodes = ICBarcode_FindBarcodes(pCustomData->pIC_BarcodeScanner,
                                            &buffer[0],
                                            width,height,width,
                                            pCustomData->BarCodes.data(), pCustomData->BarCodes.size());

            }
            else
            {
                pCustomData->FoundBarcodes = ICBarcode_FindBarcodes(pCustomData->pIC_BarcodeScanner,
                                            info.data,
                                            width,height,width,
                                            pCustomData->BarCodes.data(), pCustomData->BarCodes.size());
            }
            
            //if(pCustomData->FoundBarcodes> 0)
            pCustomData->pMainWindow->postBarcodeFoundEvent(1,2);

        }
        
        // Calling Unref is important!
        gst_buffer_unmap (buffer, &info);
        pCustomData->busy = false;
    }


    gst_sample_unref(sample);
    
    return GST_FLOW_OK;
}
